#!/usr/bin/perl -w
# cibull.pl (calcul incrmental du classement pour le jeu the bull)
# Auteur: Jean-Marie Pacquet
#
# Aujourd'hui ou pour une date donne (option "-x aaaa-mm-jj") il calcule:
# 1. le rendement de chaque portefeuille
# 2. l'cart-type
# 3. le ratio de Sharpe
# en reprenant les rsultats venant de l'xcution de ce mme programme
#  des dates antrieures
# il classe ensuite les portefeuilles pour chacun de ces trois critres
# il transfre via FTP (option "-f") les fichiers rsultants sur le serveur
#
use strict;
use Getopt::Std;
use DBI;
use Net::FTP;

#
###############################################################################
#
my $urlftp = "ftp.server.ext";
my $loginftp = "login";
my $mdpftp = "password";
my $loginpg = "login";
my $mdppg = "password";
# date du classement
my $anc;
my $moisc;
my $jourc;
# date du jour
my ($sec,$min,$heure,$jour,$mois,$an,$js,$ja,$hiv) = localtime(time);
my $ana; # anne d'avant
# Rcuprer les options de la ligne de commande
our($opt_d, $opt_x, $opt_f);
if (getopts('dfx:')) {
    print $opt_x . "\n" if ($opt_x);
} else {
    die "Usage: cibull.pl [-x aaaa-mm-jj] [-f]\n";
}
if ($opt_x) {
    $anc = substr($opt_x,0,4);
    $moisc = substr($opt_x,5,2);
    $jourc = substr($opt_x,8,2);
    $ana = $anc - 2001;
} else {
    $anc = $an + 1900;
    $moisc = 12;
    $jourc = 31;
    $ana = $an - 101; # extraction des deux derniers chiffres de l'anne d'avant
}
my $ann = $ana + 1; # l'anne en cours a un an de plus que celle d'avant
if ($ana < 10) {
    $ana = "0" . $ana;
}
if ($ann < 10) {
    $ann = "0" . $ann;
}
print "cibull.pl $ann\n";
if ($opt_d) {
    print "Date C = $anc-$moisc-$jourc\n";
}
# Ouvrir la base de donnes
#
my $dbh = DBI->connect('DBI:Pg:dbname=thebull',
		       $loginpg,
		       $mdppg,
		       {
			   AutoCommit => 0
		       }
		       ) or die "Erreur de connexion  la base: " . DBI->errstr;
# date de rfrence pour les cotes
my $origine = "20" . $ana . "-12-31";
# date de dernire cotation
my $date1;
my $an1 = "20" . $ana;
my $mois1 = "12";
my $jour1 = "31";
# compteur pour les jours de cotation
my $ccount; 
#
# prparer la liste des portefeuilles
#
my $joueursnn = "joueurs" . $ann;
my $select_joueur = $dbh->prepare("SELECT portif FROM $joueursnn");
my $joueur; # rsultat de la requte $select_joueur
my $select_portif = $dbh->prepare("SELECT portif,pseudo FROM $joueursnn");
my @joueur; # rsultat de la requte $select_portif
my %portif; # tableau de la valeur des portefeuilles par joueur
my %pseudo; # tableau des pseudos
#
# prparer la liste des valeurs pour un portefeuille
#
my $lignesnn = "lignes" . $ann;
my $select_lignes = $dbh->prepare("SELECT valeur FROM $lignesnn WHERE portif = ?");
my $valeur; # rsultat de la requte $select_lignes
#
# prparer la slection des valeurs d'un portefeuille trie par rang
#
my $valeursnn = "valeurs" . $ann;
my $select_vallib = $dbh->prepare("SELECT $lignesnn.valeur,libelle FROM $lignesnn JOIN $valeursnn ON $valeursnn.valeur = $lignesnn.valeur WHERE portif = ? ORDER BY rang");
my @vallib; # rsultat de la requte $select_vallib
#
# prparer la liste des jours de cotation depuis l'origine
#
my $cotationsnn = "cotations" . $ann;
my $select_jours = $dbh->prepare("SELECT DISTINCT jour FROM $cotationsnn WHERE jour > ? ORDER BY jour");
my $date; # rsultat de la requte $select_jours
my $reprise; # date de reprise des rsultats
#
# prparer la liste des cotations  une date donne
#
my $select_cotes = $dbh->prepare("SELECT cote,valeur FROM $cotationsnn WHERE jour = ?");
my $dividendesnn = "dividendes" . $ann;
my $select_dividende = $dbh->prepare("SELECT dividende,valeur FROM $dividendesnn WHERE jour = ?");
my @cote; # tableau de rsultat de la requte $select_cotes
my %cotedorigine; # tableau des cotes d'origine index par valeur
my %cotedujour; # tableau des cotes du jour index par valeur
my %dividende; # tableau des dividendes index par valeur
my @detachement; # tableau de rsultat de la requte $select_dividende
#
# prparer la slection et l'insertion du rsultat
#
my $resultatsnn = "resultats" . $ann;
my $select_resultat = $dbh->prepare("SELECT resultat,sigmad,sigmac FROM $resultatsnn WHERE portif = ? AND jour = ?");
my @totaux; # rsultat et cumuls partiels
my $insert_resultat = $dbh->prepare("INSERT INTO $resultatsnn (portif,jour,resultat) VALUES (?,?,?)");
my $update_resultat = $dbh->prepare("UPDATE $resultatsnn SET place = ?, sigmad = ?, sigmac = ? WHERE portif = ? AND jour = ?");

die "Impossible de prparer les requtes\n"
    unless defined $select_joueur && defined $select_portif
        && defined $select_lignes && defined $select_vallib
        && defined $select_jours
        && defined $select_cotes && defined $select_dividende
        && defined $select_resultat && $insert_resultat && $update_resultat;
#
# initialisation des cotes des valeurs
#    et des dividendes
#
$select_cotes->execute($origine);
$ccount = 0;
while (@cote = $select_cotes->fetchrow_array()) {
    $cotedorigine{$cote[1]} = $cote[0];
    $cotedujour{$cote[1]} = $cote[0];
    $dividende{$cote[1]} = 0;
    $ccount++;
}
if ($ccount > 0) {
    print $ccount . " initialisations de cotes  la date " . $origine . "\n";
}
#
# rcupration de la date des derniers rsultats enregistrs
#
my $select_date = $dbh->prepare("SELECT MAX(jour) FROM $resultatsnn");
die "Impossible de prparer la requte de date\n"
    unless defined $select_date;
$select_date->execute();
$reprise = $select_date->fetchrow_array();
$select_date->finish;
if (!$reprise) {
    $reprise = $origine;
}
print "Date de reprise= " . $reprise . "\n";
my $anr = substr($reprise,0,4);
my $moisr = substr($reprise,5,2);
my $jourr = substr($reprise,8,2);
#
# initialisation des cotes des valeurs
#    (y compris les dividendes)  la date de reprise
#
$date1 = $origine;
$ccount = 0; # nombre de jours de cotation
$select_jours->execute($origine);
while ($date = $select_jours->fetchrow_array()) {
    my $and = substr($date,0,4);
    my $moisd = substr($date,5,2);
    my $jourd = substr($date,8,2);
    if (($anr < $and) or (($anr == $and) and ($moisr < $moisd))
	or (($anr == $and) and ($moisr == $moisd) and ($jourr < $jourd))) {
	last;
    }
    # mise  jour des dividendes (pour les valeurs dont c'est le jour de dtachement)
    $select_dividende->execute($date);
    while (@detachement = $select_dividende->fetchrow_array()) {
        $dividende{$detachement[1]} += $detachement[0];
    }
    # mise  jour des cotes (pour les valeurs qui ont cot)
    $select_cotes->execute($date);
    while (@cote = $select_cotes->fetchrow_array()) {
	$cotedujour{$cote[1]} = $cote[0];
	if ($dividende{$cote[1]}) {
	    $cotedujour{$cote[1]} = sprintf("%.2f",$cote[0] + $dividende{$cote[1]});
	}
    }
    # dernire date de cotation avant la reprise
    $date1 = $date;
    # ajustement du nombre de jours de cote
    $ccount += 1;
}
print "Dernire date de cotation avant la reprise= $date1 \n";
print $ccount . " jours de cote jusqu' la reprise\n";
$select_cotes->finish;
$select_jours->finish;
#
# initialisation de la valeur des portefeuilles,
# du cumul des variations de valeur (cumuldelta)
# et du cumul des carrs de cette variation (cumulcarres)
#
my $pcount = 0;
my %cumuldelta; # cumul des variations de valeur (pour le calcul de la moyenne)
my %cumulcarres; # cumul des carrs des variations de valeur (pour le calcul de la variance)
$select_portif->execute();
while (@joueur = $select_portif->fetchrow_array()) {
    # rcupration de la valeur du portif et des cumuls au jour prcdent
    $select_resultat->execute($joueur[0],$date1);
    if ($select_resultat->rows == 0) {
	if ($date1 ne $origine) {
	    die "Rsultat prcdent manquant\n";
	} else {
	    $portif{$joueur[0]} = 100;
	    $cumuldelta{$joueur[0]} = 0;
	    $cumulcarres{$joueur[0]} = 0;
	}
    } else {
	@totaux = $select_resultat->fetchrow_array();
	$portif{$joueur[0]} = 100 + $totaux[0];
	$cumuldelta{$joueur[0]} = $totaux[1];
	$cumulcarres{$joueur[0]} = $totaux[2];
    }
    $pseudo{$joueur[0]} = "$joueur[1]";
    $pcount += 1;
}
print $pcount . " initialisations de portefeuilles\n";
#
# boucle de mise  jour du rsultat de chaque portefeuille
# (pour chaque jour de cotation depuis la date des rsultats prcdents)
# calcul du cumul des variations de valeur
# et calcul du cumul des carrs de cette variation
#
my $somme;
my $numligne;
my $variation;
my $pourcentage;
my $resultat;
my %rendt; # le rendement
my @classement; # le classement final
my $place;
my $rcount = 0; # nombre de mises  jour de rsultat
my $rstart;
$an1 = substr($reprise,0,4);
$mois1 = substr($reprise,5,2);
$jour1 = substr($reprise,8,2);
$select_jours->execute($reprise);
if ($select_jours->rows == 0) {
    print "Pas de mise  jour des rsultats\n";
} else {
  while ($date = $select_jours->fetchrow_array()) {
    print "Date= " . $date . "\n";
    my $and = substr($date,0,4);
    my $moisd = substr($date,5,2);
    my $jourd = substr($date,8,2);
    if (($anc < $and) or (($anc == $and) and ($moisc < $moisd))
	or (($anc == $and) and ($moisc == $moisd) and ($jourc < $jourd))) {
	last;
    }
    # mise  jour des dividendes (pour les valeurs dont c'est le jour de dtachement)
    $select_dividende->execute($date);
    while (@detachement = $select_dividende->fetchrow_array()) {
        $dividende{$detachement[1]} += $detachement[0];
    }
    # mise  jour des cotes (pour les valeurs qui ont cot)
    $select_cotes->execute($date);
    while (@cote = $select_cotes->fetchrow_array()) {
	$cotedujour{$cote[1]} = $cote[0];
	if ($dividende{$cote[1]}) {
	    $cotedujour{$cote[1]} = sprintf("%.2f",$cote[0] + $dividende{$cote[1]});
	}
    }
    #
    # Pour chaque joueur (et donc chaque portefeuille)
    #
    $rstart = $rcount;
    $select_joueur->execute();
    while ($joueur = $select_joueur->fetchrow_array()) {
#	print "Joueur= " . $joueur . "\n"; # DEBUG
	# est-ce qu'on a dj le rsultat?
	$select_resultat->execute($joueur,$date);
	if ($select_resultat->rows == 0) {
	    # on ne l'a pas, on le calcule
	    $select_lignes->execute($joueur);
	    $somme = 0;
	    $numligne = 0;
	    # pour chaque ligne du portefeuille
	    while ($valeur = $select_lignes->fetchrow_array()) {
		$variation = 100 * ($cotedujour{$valeur} - $cotedorigine{$valeur}) / $cotedorigine{$valeur};
		$somme += $variation;
		$numligne++;
	    }
	    $select_lignes->finish;
	    $resultat = $somme;
	    if ($numligne > 1) {
		$resultat /= $numligne;
	    }
	    # on l'insre
	    $insert_resultat->execute($joueur,$date,$resultat)
		or die "Impossible d'ajouter un rsultat\n";
	    $dbh->commit();
	    $rcount += 1;
	} else {
	    # on a dj le rsultat, on le rcupre
	    @totaux = $select_resultat->fetchrow_array();
	    $resultat = $totaux[0];
	}
	# variation en pourcentage
	$variation = ($resultat + 100 - $portif{$joueur}) / $portif{$joueur};
	# nouvelle valeur du portefeuille
	$rendt{$joueur} = $resultat;
	$portif{$joueur} = 100 + $resultat;
	# cumul des variations pour ce portefeuille
	$cumuldelta{$joueur} += $variation;
	# cumul des carrs des variations pour ce portefeuille
	$cumulcarres{$joueur} += ($variation * $variation);
    }
    if ($rcount > $rstart) {
	#
	# tri par ordre dcroissant du rendement
	#
	@classement = sort { $rendt{$b} <=> $rendt{$a} } keys %rendt;
	# enregistrement de la place du joueur dans le classement
	$place = 0;
	foreach $joueur (@classement) {
	    if ($joueur > 0) {
		$place += 1;
	    }
	    $update_resultat->execute($place,$cumuldelta{$joueur},$cumulcarres{$joueur},$joueur,$date)
		or die "Impossible d'enregistrer la place\n";
	    $dbh->commit();
	}
    }
    # dernire date de cotation
    $an1 = $and;
    $mois1 = $moisd;
    $jour1 = $jourd;
    # ajustement du nombre de jours de cote
    $ccount += 1;
  }
  print $rcount . " mises  jour de rsultats\n";
}
print $ccount . " jours de cote\n";
#
# boucle de calcul final pour chaque joueur
#
my $moyenne; # la moyenne des variations de valeur du portefeuille
my $variance; # la variance (= moyenne des carrs des variations)
my %ecart_type; # l'cart type (= racine carre de la variance)
my $fichier_etype;
my %sharpe; # le ratio de Sharpe
my $fichier_sharpe;
my $livreta = .03; # le taux de rendement sans risque (livret A)
#
$select_joueur->execute();
while ($joueur = $select_joueur->fetchrow_array()) {
    if ($ccount == 0) {
	$moyenne = 0;
	$rendt{$joueur} = 0;
	$ecart_type{$joueur} = 0;
	$sharpe{$joueur} = 1000000;
    } else {
	# calcul de la moyenne des variations
	$moyenne = $cumuldelta{$joueur} / $ccount;
#    print "Joueur n" . $joueur . "\nMoyenne=" . $moyenne; # debug
	$rendt{$joueur} = $portif{$joueur} - 100;
#    print "\tRsultat=" . $rendt{$joueur} . "\n"; # debug
	# calcul de la variance et de l'cart-type de la priode
	$variance = ($cumulcarres{$joueur} / $ccount) - ($moyenne * $moyenne);
	if ($variance <= 0) {
	    $ecart_type{$joueur} = 0;
	} else {
	    $ecart_type{$joueur} = sqrt($variance);
	}
#    print "Variance=" . $variance . "\tEcart-type=" . $ecart_type{$joueur} . "\n"; # debug
	# calcul du ratio de Sharpe
	if ($ecart_type{$joueur} != 0) {
	    $sharpe{$joueur} = (($rendt{$joueur} / 100) - $livreta) / $ecart_type{$joueur};
	} else {
	    next;
	}
#    print "Sharpe=" . $sharpe{$joueur} . "\n"; # debug
    }
}
if ($ccount == 0) {
    #
    # tri par ordre alphabtique des pseudos (listes valides)
    #
    @classement = sort { uc $pseudo{$a} cmp uc $pseudo{$b} } keys %pseudo;
} else {
    #
    # tri par ordre dcroissant du rendement
    #
    @classement = sort { $rendt{$b} <=> $rendt{$a} } keys %rendt;
}
my ($sec2,$min2,$heure2,$jour2,$mois2,$an2,$js2,$ja2,$hiv2) = localtime(time);
$an2 += 1900;
$mois2++;
$mois2 = $mois2 < 10 ? "0" . $mois2 : $mois2;
$jour2 = $jour2 < 10 ? "0" . $jour2 : $jour2;
# affichage du classement (dans un fichier en html)
my $fichier_rendt = "rendt" . substr($an1,2,2) . $mois1 . $jour1 . ".html";
if ($ccount == 0) {
    $fichier_rendt = "listes20" . $ann . ".html";
}
open(RENDT, ">" . $fichier_rendt);
if ($ccount == 0) {
    print RENDT "<h3 class=t6>Participants au " . $jour2 . "/" . $mois2 . "/" . $an2 . "</h3>\n";
} else {
    print RENDT "<h3 class=t6>Classement gnral du " . $jour1 . "/" . $mois1 . "/" . $an1 . "</h3>\n";
}
print RENDT "<h5 class=t6>(Fait le " . $jour2 . "/" . $mois2 . "/" . $an2 ."  ";
print RENDT $heure . ":" . $min . ":" . $sec . ")</h5>\n";
print RENDT "<table border=0 cellpadding=1 cellspacing=1 width=100%><tbody>";
$place = 0;
foreach $joueur (@classement) {
    if ($joueur < 0) {
	print RENDT "<tr class=t1><td>&nbsp;</td>";
    } else {
	$place += 1;
	print RENDT "<tr class=t1><td>" . $place . "</td>";
    }
    print RENDT "<td colspan=2>" . $pseudo{$joueur} . "</td>";
    if ($rendt{$joueur} < 0) {
	print RENDT "<td colspan=2 class=t3><b>";
    } else {
	print RENDT "<td colspan=2 class=t2><b>";
    }
    print RENDT sprintf("%+.2f",$rendt{$joueur}) . "</b></td></tr>\n";
    my $impair = 1;
    $select_vallib->execute($joueur);
    # pour chaque ligne du portefeuille
    while (@vallib = $select_vallib->fetchrow_array()) {
	if ($impair) {
	    print RENDT "<TR CLASS=t7><TD>&nbsp;</TD>";
	    $impair = 0;
	} else {
	    print RENDT "<TR CLASS=t8><TD BGCOLOR=WHITE>&nbsp;</TD>";
	    $impair =1;
	}
	$valeur = $vallib[0];
	print RENDT "<TD>$vallib[1]</TD>";
	print RENDT "<TD>$cotedorigine{$valeur}</TD>";
	print RENDT "<TD>$cotedujour{$valeur}</TD>";
        $variation = 100 * ($cotedujour{$valeur} - $cotedorigine{$valeur}) / $cotedorigine{$valeur};
        $pourcentage = sprintf("%+01.2f%%",$variation);
        if ($variation >= 0) {
            print RENDT "<TD CLASS=t4>$pourcentage</TD></TR>\n";
        } else {
            print RENDT "<TD CLASS=t5>$pourcentage</TD></TR>\n";
        }
    }
    print RENDT "\n";
}
$select_vallib->finish;
print RENDT "</tbody></table>\n<br>\n";
close(RENDT);
# A partir de fvrier
if (($ccount > 0) and ($mois1 > 1)) {
    #
    # tri par ordre croissant de l'cart type
    #
    @classement = sort { $ecart_type{$a} <=> $ecart_type{$b} } keys %sharpe;
    ($sec,$min,$heure,$jour2,$mois2,$an2,$js,$ja,$hiv) = localtime(time);
    $an2 += 1900;
    $mois2++;
    $mois2 = $mois2 < 10 ? "0" . $mois2 : $mois2;
    $jour2 = $jour2 < 10 ? "0" . $jour2 : $jour2;
    # affichage du classement (dans un fichier en html)
    $fichier_etype = "etype" . substr($an1,2,2) . $mois1 . $jour1 . ".html";
    open(ECARTTYPE, ">" . $fichier_etype);
    print ECARTTYPE "<h3 class=t6>Classement cart-type du " . $jour1 . "/" . $mois1 . "/" . $an1 . "</h3>\n";
    print ECARTTYPE "<h5 class=t6>(Fait le " . $jour2 . "/" . $mois2 . "/" . $an2 . "  ";
    print ECARTTYPE $heure . ":" . $min . ":" . $sec . ")</h5>\n";
    print ECARTTYPE "<table border=0 cellpadding=1 cellspacing=1 width=100%><tbody>";
    $place = 0;
    foreach $joueur (@classement) {
	if ($joueur < 0) {
	    print ECARTTYPE "<tr class=t1><td>&nbsp;</td>";
	} else {
	    $place += 1;
	    print ECARTTYPE "<tr class=t1><td>" . $place . "</td>";
	}
	print ECARTTYPE "<td colspan=2>" . $pseudo{$joueur} . "</td>";
	if ($sharpe{$joueur} < 0) {
	    print ECARTTYPE "<td colspan=2 class=t3><b>";
	} else {
	    print ECARTTYPE "<td colspan=2 class=t2><b>";
	}
	print ECARTTYPE sprintf("%+.5f",$ecart_type{$joueur}) . "</b></td></tr>\n";
    }
    print ECARTTYPE "</tbody></table>\n<br>\n";
    close(ECARTTYPE);
    #
    # tri par ordre dcroissant du ratio de Sharpe
    #
    @classement = sort { $sharpe{$b} <=> $sharpe{$a} } keys %sharpe;
    ($sec,$min,$heure,$jour2,$mois2,$an2,$js,$ja,$hiv) = localtime(time);
    $an2 += 1900;
    $mois2++;
    $mois2 = $mois2 < 10 ? "0" . $mois2 : $mois2;
    $jour2 = $jour2 < 10 ? "0" . $jour2 : $jour2;
    # affichage du classement (dans un fichier en html)
    $fichier_sharpe = "sharpe" . substr($an1,2,2) . $mois1 . $jour1 . ".html";
    open(SHARPE, ">" . $fichier_sharpe);
    print SHARPE "<h3 class=t6>Classement Sharpe du " . $jour1 . "/" . $mois1 . "/" . $an1 . "</h3>\n";
    print SHARPE "<h5 class=t6>(Fait le " . $jour2 . "/" . $mois2 . "/" . $an2 . "  ";
    print SHARPE $heure . ":" . $min . ":" . $sec . ")</h5>\n";
    print SHARPE "<table border=0 cellpadding=1 cellspacing=1 width=100%><tbody>";
    $place = 0;
    foreach $joueur (@classement) {
	if ($joueur < 0) {
	    print SHARPE "<tr class=t1><td>&nbsp;</td>";
	} else {
	    $place += 1;
	    print SHARPE "<tr class=t1><td>" . $place . "</td>";
	}
	print SHARPE "<td colspan=2>" . $pseudo{$joueur} . "</td>";
	if ($sharpe{$joueur} < 0) {
	    print SHARPE "<td colspan=2 class=t3><b>";
	} else {
	    print SHARPE "<td colspan=2 class=t2><b>";
	}
	print SHARPE sprintf("%+.2f",$sharpe{$joueur}) . "</b></td></tr>\n";
    }
    print SHARPE "</tbody></table>\n<br>\n";
    close(SHARPE);
}
$select_resultat->finish;
$select_joueur->finish;
$select_cotes->finish;
$select_jours->finish;
#
# fin de la boucle, on ferme!
#
$dbh->disconnect;
#
# Transfert des fichiers html vers le serveur (option -f)
#
if ($opt_f) {
    my $ftp = new Net::FTP($urlftp, Debug => 0, Passive => 1);
    $ftp->login($loginftp,$mdpftp);
    $ftp->cwd("/thebull/20" . $ann);
    $ftp->put($fichier_rendt);
    if (($ccount > 0) and ($mois1 > 1)) {
	$ftp->put($fichier_etype);
	$ftp->put($fichier_sharpe);
    }
    $ftp->quit;
}
# c'est fini!
exit(0);


