#!/usr/bin/perl -w
# clvexx.pl
# Classement des vétérans sur une période donnée
# A défaut de période, classement sur la durée du jeu
#
# Ceux qui n'ont pas participé à un maximum de un quart des éditions reçoivent
# pour le calcul la performance moyenne des joueurs de l'édition concernée
# En option (bimbo) c'est zéro.
#
# Auteur: Jean-Marie Pacquet
#
#
use strict;
use Getopt::Std;
use DBI;
use Net::FTP;
use utf8;
#
my ($sec,$min,$heure,$jour,$mois,$an,$js,$ja,$hiv) = localtime(time);
$an += 1900;
$mois++;
$mois = $mois < 10 ? "0" . $mois : $mois;
$jour = $jour < 10 ? "0" . $jour : $jour;
my $andhui = $an;
my $ann = $an - 2000; # édition du jeu = année sur deux chiffres
print "clvexx.pl $ann\n";
#
###############################################################################
#
# Récupérer les options de la ligne de commande
#
# a=append (ajout au fichier s'il existe)
# b=bimbo (zéro pour les années manquantes, par défaut la moyenne des joueurs)
# d=debug (imprime les étapes du calcul)
# f=from (année de début de la période, par défaut 2004, la première année du jeu)
# m=missed (nombre d'années manquantes maximum, par défaut aucune)
# t=to (année de fin de la période, par défaut l'année en cours))
# x=xfer (transfert au site du fichier de classement)
#
our($opt_a, $opt_b, $opt_d, $opt_f, $opt_m, $opt_t, $opt_x);
if (getopts('abdf:m:t:x')) {
    print "Version bimbo (zéro pour les éditions manquées)\n" if ($opt_b);
    if (!$opt_f) {
	$opt_f = 2004;
    }
    print "A partir de $opt_f\n";
} else {
    die "Usage: clvexx.pl [-[b][d][x]] [-m nombre] [-f yyyy] [-t yyyy]\n";
}
#
my $loginpg = "login";
my $mdppg = "password";
# Ouvrir la base de données
#
my $dbh = DBI->connect('DBI:Pg:dbname=thebull',
		       $loginpg,
		       $mdppg,
		       {
			   AutoCommit => 0
		       }
		       ) or die "Erreur de connexion à la base: " . DBI->errstr;
# Aller chercher la moyenne de l'année dans la table des résultats
my $rtable = 'resultats';
#
# Aller chercher la liste des portefeuilles des joueurs dans la table des participations
#
my $jtable = 'joueursxx';
my $editions; # nombre d'editions du jeu
my $partmini; # minimum de participations pour le nombre d'editions
#
# préparer le calcul de la moyenne et la recherche des joueurs
#
my $select_moyenne = $dbh->prepare('SELECT resultat FROM resultats WHERE portif = 0');
my %moyenne; # la moyenne des résultats par année
my $select_resultat = $dbh->prepare('SELECT resultat FROM resultats WHERE portif = ? and jour = ?');
my $resultat; # résultat
my $date; # date du résultat
my $reveillon;
my $select_participations = $dbh->prepare('SELECT portif, pseudo, an FROM joueursxx WHERE portif IN (SELECT portif FROM joueursxx WHERE an >= ? and an <= ? GROUP BY portif HAVING count(an) >= ?) ORDER BY portif, an DESC');
my @participation; # participation à une édition
my $portif;
my %pseudo; # les pseudos des joueurs
my %pxx; # le tableau des performances
my $ctable = 'cotations' . $ann;
my $select_date = $dbh->prepare("SELECT MAX(jour) from $ctable");

die "Impossible de préparer les requêtes\n"
    unless defined $select_moyenne && defined $select_resultat
    && defined $select_participations && defined $select_date;

if (!$opt_t) {
    $opt_t = $andhui;
}
print "Jusqu'à $opt_t\n";
# Boucle sur toutes les éditions du JTB pour trouver (ou calculer) la moyenne
$editions = 0;
for (my $an = 2004; $an <= $andhui; $an++) {
    # Ajuster les paramètres en fonction de l'année
    $date = $an . '-12-31';
    if ($an > 2004) {
	my $an2 = substr($date,2,2);
	substr($rtable,-2,2) = $an2;
	if ($opt_d) {
	    print "Table pour $an = $rtable\n";
	}
	if ($an >= $andhui) {
	    $select_date->execute();
	    if ($select_date->rows == 0) {
		die "Pas de cotation en $andhui\n";
	    }
	    $date = $select_date->fetchrow_array();
	    if ($opt_d) {
		print "Dernier classement le $date ($ja e jour) \n";
	    }
	    $select_date->finish;
	}
    }
    # Lire la moyenne de l'année (si on l'a déjà calculée)
    $select_moyenne = $dbh->prepare("SELECT resultat FROM $rtable WHERE portif = 0 AND jour = ?")
	or die "Impossible de préparer la lecture de la moyenne\n";
    $select_moyenne->execute($date);
    if ($select_moyenne->rows == 0) {
	# Calculer la moyenne pour l'année
	$select_moyenne = $dbh->prepare("SELECT AVG(resultat) FROM $rtable WHERE portif > 0 AND jour = ?")
	    or die "Impossible de préparer le calcul de la moyenne\n";
	$select_moyenne->execute($date);
	$moyenne{$an} = $select_moyenne->fetchrow_array();
	# Inscrire la moyenne dans la table des résultats des années précédentes
	if ($an < $andhui) {
	    $select_moyenne = $dbh->prepare("INSERT INTO $rtable (portif, resultat, jour) VALUES (0, $moyenne{$an}, '$date')")
		or die "Impossible de préparer l'écriture de la moyenne\n";
	    $select_moyenne->execute();
	}
    } else {
	$moyenne{$an} = $select_moyenne->fetchrow_array();
    }
    if ($opt_d) {
	print "Moyenne $an = $moyenne{$an} \n";
    }
   # préparation des nouvelles requêtes
    if ($an == 2004) {
	$rtable .= '04';
    }
    $editions++;
}
$select_moyenne->finish;
$dbh->commit();
$ann = substr($date,2,2); # edition du jeu "11" pour "2011"
my $an1 = substr($date,0,4);
my $mois1 = substr($date,5,2);
my $jour1 = substr($date,8,2);
#
# Boucle de calcul de performance pour chaque portif remplissant les conditions
#
$editions = $opt_t - $opt_f + 1;
if ($opt_m) {
    $partmini = $editions - $opt_m;
} else {
#    $partmini = int((($editions * 3) + 3) / 4);
    $partmini = $editions;
}
if ($opt_d) {
    print "Participations requises = $partmini pour $editions editions\n";
}
$select_participations->execute($opt_f, $opt_t, $partmini);
$portif = 0;
$an = $opt_t;
while ( @participation = $select_participations->fetchrow_array() ) {
    if (($participation[2] < $opt_f) || ($participation[2] > $opt_t)) {
	next;
    }
    if ($participation[0] != $portif) {
	if ($portif != 0) {
	    # nouveau participant, est-ce fini avec le précédent?
	    while ($an >= $opt_f) {
		# Non, combler le trou de début de période
		# récupérer la perf de substitution
		if ($opt_b) {
		    $resultat = 0; # version bimbo
		} else {
		    $resultat = $moyenne{$an};
		}
		# Mettre à jour la performance sur la durée
		$pxx{$portif} = $pxx{$portif} + $resultat + (($pxx{$portif} * $resultat) / 100);
		# Ajouter une marque d'absence au pseudo
		$pseudo{$portif} .= '*';
		if ($opt_d) {
		    print "\n$an manqué\n";
		}
		$an--;
	    }
	    # affichage de la performance calculée
	    if ($opt_d) {
		print "\n$pseudo{$portif} ($portif) Perf=$pxx{$portif}\n";
	    }
	    # reinitialiser l'année sur la dernière année du classement
	    $an = $opt_t;
	}
	# initialiser le portif et le pseudo du participant
	$portif = $participation[0];
	$pseudo{$portif} = "$participation[1]";
    }
    if ($opt_d) {
	print "$an: $participation[1] ($participation[0]) $participation[2]\t";
    }
    if ($participation[2] < $an) {
	while ($participation[2] < $an) {
	    # pas de participation pour cette année là
	    # récupérer la perf de substitution
	    if ($opt_b) {
		$resultat = 0; # version bimbo
	    } else {
		$resultat = $moyenne{$an};
	    }
	    if ($an == $opt_t) {
		# Initialiser la performance au pro rata de l'année en cours si nécessaire
		if ($an == $andhui) {
		    $pxx{$portif} = ($resultat * $ja ) / 365;
		} else {
		    $pxx{$portif} = $resultat;
		}
	    } else {
		# Mettre à jour la performance sur la durée
		$pxx{$portif} = $pxx{$portif} + $resultat + (($pxx{$portif} * $resultat) / 100);
	    }
	    # Ajouter une marque d'absence au pseudo
	    $pseudo{$portif} .= '*';
	    if ($opt_d) {
		print "\nPerf=$pxx{$portif}\n$an manqué\n";
	    }
	    $an--;
	}
    }
    if ($an == $andhui) {
	$rtable = 'resultats' . substr($date,2,2);
	# Aller chercher le résultat en cours pour ce portif
	$select_resultat = $dbh->prepare("SELECT resultat from $rtable where portif = $portif and jour = '$date'");
	$select_resultat->execute();
	if ($select_resultat->rows == 0) {
	    print "Pas de résultat dans $rtable pour portif = $portif et jour = '$date'\n";
	    last;
	} else {
	    $resultat = $select_resultat->fetchrow_array();
	}
	# Initialiser la performance au pro rata de l'année en cours
	$pxx{$portif} = ($resultat * $ja ) / 365;
    } else {
	$reveillon = $an . '-12-31';
	if ($an > 2004) {
	    $rtable = 'resultats' . substr($reveillon,2,2);
	} else {
	    $rtable = 'resultats';
	}
	# Aller chercher le résultat de cette année là pour ce portif
	$select_resultat = $dbh->prepare("SELECT resultat FROM $rtable WHERE portif = $portif AND jour = '$reveillon'");
	$select_resultat->execute();
	if ($select_resultat->rows == 0) {
	    print "Pas de résultat dans $rtable pour portif = $portif et jour = '$reveillon'\n";
	    last;
	} else {
	    $resultat = $select_resultat->fetchrow_array();
	}
	if ($an == $opt_t) {
	    $pxx{$portif} = $resultat;
	} else {
	    # Mettre à jour la performance sur la durée
	    $pxx{$portif} = $pxx{$portif} + $resultat + (($pxx{$portif} * $resultat) / 100);
	}
    }
    if ($opt_d) {
	print "Perf=$pxx{$portif}\n";
    }
    $an--;
}
if ($portif != 0) {
    # Est-ce fini avec le précédent?
    while ($an >= $opt_f) {
	# Non, combler le trou
	# récupérer la perf de substitution
	if ($opt_b) {
	    $resultat = 0; # version bimbo
	} else {
	    $resultat = $moyenne{$an};
	}
	# Mettre à jour la performance sur la durée
	$pxx{$portif} = $pxx{$portif} + $resultat + (($pxx{$portif} * $resultat) / 100);
	# Ajouter une marque d'absence au pseudo
	$pseudo{$portif} .= '*';
	$an--;
    }
    # affichage de la performance calculée
    if ($opt_d) {
	print "\n$pseudo{$portif} ($portif) Perf=$pxx{$portif}\n";
    }
}
$select_resultat->finish;
$select_participations->finish;
#
# tri par ordre décroissant de la performance
#
my @classement;
# classement final
@classement = sort { $pxx{$b} <=> $pxx{$a} } keys %pxx;
# affichage du classement (dans un fichier en html)
my $fichier_cumul = "cumul". substr($an1,2,2) . $mois1 . $jour1 . ".html";
open(CUMUL, ($opt_a ? ">>" : ">") . $fichier_cumul);
print CUMUL "<h3 class=t6>Classement de la période $opt_f - $opt_t au " . $jour1 . "/" . $mois1 . "/" . $an1 . "</h3>\n";
if ($opt_m) {
    print CUMUL "<h4>Au maximum $opt_m années manquantes remplacées par la moyenne des joueurs pour les années correspondantes</h4>";
} else {
    print CUMUL "<h4>Sans année manquante</h4>";
}
print CUMUL "<h5 class=t6>(Fait le " . $jour . "/" . $mois . "/" . $andhui . " à ";
print CUMUL $heure . ":" . $min . ":" . $sec . ")</h5>\n";
print CUMUL "<table border=0 cellpadding=1 cellspacing=1 width=100%><tbody>";
my $place = 0;
foreach $portif (@classement) {
    if ($portif < 0) {
	print CUMUL "<tr class=t1><td>&nbsp;</td>";
    } else {
	$place += 1;
	print CUMUL "<tr class=t1><td>" . $place . "</td>";
    }
    print CUMUL "<td colspan=2>" . $pseudo{$portif} . "</td>";
    if ($pxx{$portif} < 0) {
	print CUMUL "<td class=t3><b>";
    } else {
	print CUMUL "<td class=t2><b>";
    }
    print CUMUL sprintf("%+.2f",$pxx{$portif}) . "</b></td>";
    print CUMUL "</tr>\n";
}
print CUMUL "</tbody></table>\n<br>\n";
close(CUMUL);
#
# fin des boucles, on ferme!
#
$dbh->disconnect;
#
# Transfert du fichier html vers le serveur (option -x)
#
if ($opt_x) {
    my $ftp = new Net::FTP("ftp.server.ext", Debug => 0, Passive => 1);
    $ftp->login("login","password");
    $ftp->cwd("/thebull/20" . $ann);
    $ftp->put($fichier_cumul);
    $ftp->quit;
}
# c'est fini!
exit(0);


